from datetime import timedelta

from django.contrib.auth import get_user_model
from django.test import TestCase, RequestFactory
from django.utils import timezone

from django_login_history2.app_settings import get_geolocation_helper_class
from django_login_history2.helper import IPCheckerTestMode
from django_login_history2.models import Login


class LoginHistoryTestCase(TestCase):
    def setUp(self):
        self.user = get_user_model().objects.create_user(username='testuser', password='testpass')
        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.request.user = self.user

    def test_class_helper(self):
        self.assertEqual(get_geolocation_helper_class(), IPCheckerTestMode)
        obj = get_geolocation_helper_class()(self.request, self.user)
        self.assertEqual(obj.user, self.request.user)
        self.assertEqual(obj.client_ip, '127.0.0.1')

    def test_user_login_logout(self):
        self.client.login(username='testuser', password='testpass')
        self.assertTrue(Login.objects.count() > 0)

        obj = self.user.login_history.last()
        str_login = str(obj)
        login_ts = timezone.now() - obj.login_at

        self.assertTrue(login_ts <= timedelta(seconds=5))
        self.client.logout()

        obj.refresh_from_db()
        str_logout = str(obj)

        self.assertNotEqual(str_login, str_logout)
        logout_ts = timezone.now() - obj.logout_at
        self.assertTrue(logout_ts <= timedelta(seconds=5))


import importlib
import sys
from typing import Type
from django.conf import settings
from django.core.cache import caches
from django.core.cache.backends.base import BaseCache
from django_login_history2.helper import IPCheckerAbstract

IS_TESTING = sys.argv[1:2] == ['test']

GEOLOCATION_METHOD = getattr(
    settings,
    'LOGIN_HISTORY_GEOLOCATION_METHOD',
    'django_login_history2.utils.get_geolocation_data'
)

GEOLOCATION_HELPER_CLASS = getattr(
    settings,
    'LOGIN_HISTORY_GEOLOCATION_HELPER_CLASS',
    'django_login_history2.helper.IPCheckerIPApi'
)

def get_geolocation_helper_class() -> Type[IPCheckerAbstract]:
    module_name, class_name = GEOLOCATION_HELPER_CLASS.rsplit('.', 1)
    try:

        if IS_TESTING:
            class_name = 'IPCheckerTestMode'

        module = importlib.import_module(module_name)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as er:
        raise ValueError("Invalid geolocation class specified in settings.\n", er) from er



GEOLOCATION_BLOCK_FIELDS = (
    'user',
    'ip',
    'user_agent',
    'ip_info',
    'created_at'
)

GEOLOCATION_PLACEHOLDER_IP = getattr(
    settings,
    'LOGIN_HISTORY_GEOLOCATION_PLACEHOLDER_IP',
    '8.8.8.8'
)

CACHE = getattr(
    settings,
    'LOGIN_HISTORY_GEOLOCATION_CACHE',
    'default'
)

CACHE_TIMEOUT = getattr(
    settings,
    'LOGIN_HISTORY_GEOLOCATION_CACHE_TIMEOUT',
    3600
)

def get_cache() -> BaseCache:
    return caches[CACHE]

DEBUG = getattr(settings, 'DEBUG', True)



import abc

import requests
from django.contrib.auth.models import User
from django.http import HttpRequest
from ipware import get_client_ip

from django_login_history2.dto import IPInfo


class IPCheckerAbstract(abc.ABC):
    __request: HttpRequest
    __client_ip: str
    __is_routable: bool
    __user : User
    def __init__(self, request: HttpRequest, user: User):
        self.__request = request
        self.__client_ip, self.__is_routable = get_client_ip(request)
        self.__user = user

    @property
    def client_ip(self) -> str:
        return self.__client_ip

    @property
    def is_routable(self) -> bool:
        return self.__is_routable

    @property
    def user_agent(self) -> str:
        return self.__request.META.get('HTTP_USER_AGENT', '')

    @property
    def user(self):
        return self.__user

    @abc.abstractmethod
    def get_geolocation_data(self) -> IPInfo:
        return IPInfo(
            user=self.user.pk,
            ip=self.client_ip,
            user_agent=self.user_agent,
        )


class IPCheckerTestMode(IPCheckerAbstract):

    def get_geolocation_data(self) -> IPInfo:
        from django_login_history2.app_settings import GEOLOCATION_PLACEHOLDER_IP
        return super().get_geolocation_data().with_overrides(
            ip=GEOLOCATION_PLACEHOLDER_IP,
            network='8.8.8.0/24',
            version='IPv4',
            city='Mountain View',
            region='California',
            region_code='CA',
            country='US',
            country_code='US',
            country_name='United States',
            currency='USD',
            country_code_iso3='USA',
            country_capital='Washington',
            country_tld='.us',
            continent_code='NA',
            in_eu=False,
            postal='94043',
            latitude=37.42301,
            longitude=-122.083352,
            timezone='America/Los_Angeles',
            utc_offset="-0800",
            country_calling_code='+1',
            currency_name='Dollar',
            languages='en-US,es-US,haw,fr',
            country_area=9629091.0,
            country_population=327167434,
            asn='AS15169',
            org='GOOGLE'
        )


class IPCheckerIPApi(IPCheckerAbstract):

    def get_geolocation_data(self) -> IPInfo:
        from django_login_history2.app_settings import get_cache, CACHE_TIMEOUT
        key = f'ipapi:{self.client_ip}'
        data = get_cache().get(key)

        if not data:
            data = super().get_geolocation_data()
            if not self.is_routable:
                data = data.with_overrides(error=True, reason="Address not routable")
            else:
                with requests.get(f'https://ipapi.co/{self.client_ip}/json/', timeout=60) as handler:
                    data = data.with_overrides(**handler.json())

            get_cache().set(key, data, timeout=CACHE_TIMEOUT)

        return data



import importlib
from dataclasses import asdict
from django.contrib.auth import get_user_model
from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.core.exceptions import FieldDoesNotExist
from django.dispatch import receiver
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext as _
from .dto import IPInfo
from .app_settings import (GEOLOCATION_METHOD, GEOLOCATION_BLOCK_FIELDS)


class LoginQuerySet(models.QuerySet):
    def filter(self, *args, **kwargs):
        if "login_at" in kwargs:
            kwargs["created_at"] = kwargs.pop("login_at")
        return super().filter(*args, **kwargs)

    def order_by(self, *fields):
        new_fields = [field if field != "login_at" else "created_at" for field in fields]
        return super().order_by(*new_fields)


class LoginManager(models.Manager):
    def get_queryset(self):
        return LoginQuerySet(self.model, using=self._db)


class Login(models.Model):
    user = models.ForeignKey(get_user_model(),
                             on_delete=models.CASCADE,
                             db_index=True,
                             verbose_name=_("user"),
                             related_name='login_history')
    ip = models.GenericIPAddressField(db_index=True, verbose_name=_("ip address"))
    ip_info = models.JSONField(default=dict, verbose_name=_("ip info"))
    city = models.CharField(max_length=80, blank=True, verbose_name=_("city"))
    region = models.CharField(max_length=80, blank=True, verbose_name=_("region"))
    region_code = models.CharField(max_length=10, blank=True, verbose_name=_("region code"))
    country_code = models.CharField(max_length=2, blank=True, verbose_name=_("country code"))
    country_name = models.CharField(max_length=80, blank=True, verbose_name=_("country name"))
    currency = models.CharField(max_length=5, blank=True, verbose_name=_("currency"))
    user_agent = models.TextField(verbose_name=_("user agent"))
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_("created at"), db_index=True)
    logout_at = models.DateTimeField(null=True, blank=True, verbose_name=_("logout at"), db_index=True)
    objects = LoginManager()

    @property
    def login_at(self):
        return self.created_at

    def __str__(self):
        user = getattr(self.user, "username", self.ip)
        logout = ''

        if self.logout_at:
            logout = f' and logout at {self.logout_at}'

        return f"{user} logged in at {self.created_at}{logout}"


@receiver(user_logged_in)
def post_login(sender, user, request, **kwargs):
    method_path = GEOLOCATION_METHOD
    result = None
    module_name, func_name = method_path.rsplit('.', 1)

    try:
        module = importlib.import_module(module_name)
        geolocation_func = getattr(module, func_name)
        result = geolocation_func(request, user)
    except (ImportError, AttributeError) as er:
        raise ValueError("Invalid geolocation method specified in settings.\n", er) from er

    if result:
        assert isinstance(result, IPInfo)
        mapped_fields = {}
        result_dict = asdict(result)

        for key, value in result_dict.items():
            if key in GEOLOCATION_BLOCK_FIELDS:
                continue

            try:
                _ = Login._meta.get_field(key)
                mapped_fields[key] = value
            except FieldDoesNotExist:
                pass

        _ = Login.objects.create(
            user=user,
            ip=result.ip,
            user_agent=result.user_agent,
            ip_info=result_dict,
            **mapped_fields
        )


@receiver(user_logged_out)
def post_logout(sender, request, user, **kwargs):
    last_login = Login.objects.filter(user=user).latest("created_at")
    last_login.logout_at = timezone.now()
    last_login.save(update_fields=["logout_at"])



# Generated by Django 5.1.6 on 2025-02-23 16:57

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_login_history2', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='login',
            name='logout_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='logout at'),
        ),
        migrations.AlterField(
            model_name='login',
            name='city',
            field=models.CharField(blank=True, max_length=80, verbose_name='city'),
        ),
        migrations.AlterField(
            model_name='login',
            name='country_code',
            field=models.CharField(blank=True, max_length=2, verbose_name='country code'),
        ),
        migrations.AlterField(
            model_name='login',
            name='country_name',
            field=models.CharField(blank=True, max_length=80, verbose_name='country name'),
        ),
        migrations.AlterField(
            model_name='login',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='login',
            name='currency',
            field=models.CharField(blank=True, max_length=5, verbose_name='currency'),
        ),
        migrations.AlterField(
            model_name='login',
            name='ip',
            field=models.GenericIPAddressField(db_index=True, verbose_name='ip address'),
        ),
        migrations.AlterField(
            model_name='login',
            name='ip_info',
            field=models.JSONField(default=dict, verbose_name='ip info'),
        ),
        migrations.AlterField(
            model_name='login',
            name='region',
            field=models.CharField(blank=True, max_length=80, verbose_name='region'),
        ),
        migrations.AlterField(
            model_name='login',
            name='region_code',
            field=models.CharField(blank=True, max_length=10, verbose_name='region code'),
        ),
        migrations.AlterField(
            model_name='login',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='user'),
        ),
        migrations.AlterField(
            model_name='login',
            name='user_agent',
            field=models.TextField(verbose_name='user agent'),
        ),
    ]



from django.contrib.auth.models import User
from django.http import HttpRequest
from django_login_history2.app_settings import get_geolocation_helper_class
from django_login_history2.dto import IPInfo


def get_geolocation_data(request: HttpRequest, user: User) -> IPInfo:
    cls = get_geolocation_helper_class()
    return cls(request, user).get_geolocation_data()
